{-# LANGUAGE FlexibleInstances,
             Rank2Types,
             ExistentialQuantification,
             KindSignatures,
             TypeOperators
  #-}

module AOP.Semantics.Membranes.MembraneGraph where

import Data.Graph.Inductive.Graph
import Data.Graph.Inductive.PatriciaTree
import Data.Graph.Inductive.Dot
import AOP.Internal.Joinpoint
import AOP.Internal.AOTType
import AOP.Internal.Advice
import Control.Monad.Views

type MembraneId = Node
newtype MembraneT m a = MembraneT {runMG :: (MembraneId, MembraneGraph) -> m (a, (MembraneId, MembraneGraph))}
data Membrane m n t = (Monad m, Monad n) => 
                  Membrane { mbId  :: MembraneId,
                             jpIn  :: forall a b. Jp (AOT (MembraneT m)) a b -> Maybe (Jp (AOT (MembraneT m)) a b),
                             jpOut :: forall a b. Jp (AOT (MembraneT m)) a b -> Maybe (Jp (AOT (MembraneT m)) a b),
                             -- | Tag for control flow combinator required in aspects.
                             --   The valid options are the empty data constructors:
                             --   CombinationAdvice, ReplacementAdvice, AugmentationAdvice, NarrowingAdvice
                             mbTag   :: t,     
                             mbView  :: n :><: AOT (MembraneT m)
                           }

data EMembrane = forall m n t. (Monad m, Monad n) => EMembrane (Membrane m n t)
type MembraneGraph = Gr (EMembrane) ()

emptyMembraneGraph :: MembraneGraph
emptyMembraneGraph = empty

makeMembrane :: (Monad m, Monad n) => MembraneGraph -> 
                (forall a b. Jp (AOT (MembraneT m)) a b -> Maybe (Jp (AOT (MembraneT m)) a b)) ->
                (forall a b. Jp (AOT (MembraneT m)) a b -> Maybe (Jp (AOT (MembraneT m)) a b)) ->
                t ->
                n :><: AOT (MembraneT m) ->
                Membrane m n t
makeMembrane graph liftIn liftOut comb view = let mbId = head $ newNodes 1 graph
                                    in  Membrane mbId liftIn liftOut comb view

addMembrane :: (Monad m, Monad n) => Membrane m n t -> MembraneGraph -> MembraneGraph
addMembrane membrane graph = insNode (mbId membrane, EMembrane membrane) graph

bindMembranes :: (MembraneId, MembraneId) -> MembraneGraph -> MembraneGraph
bindMembranes (mbId1, mbId2) graph = insEdge (mbId1, mbId2, ()) graph

instance Show (Membrane m n t) where
         show (Membrane id _  _ _ _) = "Mb(" ++ show id ++ ")"

instance Show EMembrane where
         show (EMembrane m) = show m

instance Graph gr => Show (gr EMembrane ()) where
        show gr = showDot (fglToDot gr)